@@ -1,134 +1,121 @@
-import * as Process from 'child_process'
-import * as Path from 'path'
+import { spawnSync } from 'child_process'
+import * as path from 'path'
 import * as TurndownService from 'turndown'
-import * as LSP from 'vscode-languageserver'
+import { isDeepStrictEqual } from 'util'
+import * as LSP from 'vscode-languageserver/node'
+import { CodeAction } from 'vscode-languageserver/node'
 import { TextDocument } from 'vscode-languageserver-textdocument'
 
 import Analyzer from './analyser'
 import * as Builtins from './builtins'
 import * as config from './config'
 import Executables from './executables'
-import { Linter } from './linter'
 import { initializeParser } from './parser'
 import * as ReservedWords from './reservedWords'
+import { Linter } from './shellcheck'
 import { BashCompletionItem, CompletionItemDataType } from './types'
 import { uniqueBasedOnHash } from './util/array'
 import { getShellDocumentation } from './util/sh'
 
 const PARAMETER_EXPANSION_PREFIXES = new Set(['$', '${'])
+const CONFIGURATION_SECTION = 'bashIde'
 
 /**
  * The BashServer glues together the separate components to implement
  * the various parts of the Language Server Protocol.
  */
 export default class BashServer {
+  private analyzer: Analyzer
+  private clientCapabilities: LSP.ClientCapabilities
+  private config: config.Config
+  private connection: LSP.Connection
+  private documents: LSP.TextDocuments<TextDocument> = new LSP.TextDocuments(TextDocument)
+  private executables: Executables
+  private linter?: Linter
+  private workspaceFolder: string | null
+  private uriToCodeActions: { [uri: string]: CodeAction[] | undefined } = {}
+
+  private constructor({
+    analyzer,
+    capabilities,
+    connection,
+    executables,
+    linter,
+    workspaceFolder,
+  }: {
+    analyzer: Analyzer
+    capabilities: LSP.ClientCapabilities
+    connection: LSP.Connection
+    executables: Executables
+    linter?: Linter
+    workspaceFolder: string | null
+  }) {
+    this.analyzer = analyzer
+    this.clientCapabilities = capabilities
+    this.connection = connection
+    this.executables = executables
+    this.linter = linter
+    this.workspaceFolder = workspaceFolder
+    this.config = {} as any // NOTE: configured in updateConfiguration
+    this.updateConfiguration(config.getDefaultConfiguration())
+  }
   /**
    * Initialize the server based on a connection to the client and the protocols
    * initialization parameters.
    */
   public static async initialize(
     connection: LSP.Connection,
-    { rootPath, capabilities }: LSP.InitializeParams,
-  ): Promise<BashServer> {
+    { rootPath, rootUri, capabilities }: LSP.InitializeParams,
+  ): // TODO: use workspaceFolders instead of rootPath
+  Promise<BashServer> {
     const { PATH } = process.env
+    const workspaceFolder = rootUri || rootPath || null
 
     if (!PATH) {
       throw new Error('Expected PATH environment variable to be set')
     }
 
     const parser = await initializeParser()
-    const analyzer = new Analyzer({ console: connection.console, parser })
-
-    let backgroundAnalysisCompleted
-    if (rootPath) {
-      // NOTE: we do not block the server initialization on this background analysis.
-      backgroundAnalysisCompleted = analyzer.initiateBackgroundAnalysis({ rootPath })
-    }
-
-    const shellcheckPath = config.getShellcheckPath()
-    const linter = new Linter({
+    const analyzer = new Analyzer({
       console: connection.console,
-      executablePath: shellcheckPath,
+      parser,
+      workspaceFolder,
     })
 
-    if (!shellcheckPath) {
-      connection.console.info('ShellCheck linting is disabled.')
-    }
-
     const executables = await Executables.fromPath(PATH)
 
     return new BashServer({
       analyzer,
-      backgroundAnalysisCompleted,
       capabilities,
       connection,
       executables,
-      linter,
+      workspaceFolder,
     })
   }
 
-  private executables: Executables
-  private analyzer: Analyzer
-  private linter: Linter
-
-  private documents: LSP.TextDocuments<TextDocument> = new LSP.TextDocuments(TextDocument)
-  private connection: LSP.Connection
-  private clientCapabilities: LSP.ClientCapabilities
-  public backgroundAnalysisCompleted?: Promise<any>
-
-  private constructor({
-    connection,
-    executables,
-    analyzer,
-    linter,
-    capabilities,
-    backgroundAnalysisCompleted,
-  }: {
-    connection: LSP.Connection
-    executables: Executables
-    analyzer: Analyzer
-    linter: Linter
-    capabilities: LSP.ClientCapabilities
-    backgroundAnalysisCompleted?: Promise<any>
-  }) {
-    this.connection = connection
-    this.executables = executables
-    this.analyzer = analyzer
-    this.linter = linter
-    this.clientCapabilities = capabilities
-    this.backgroundAnalysisCompleted = backgroundAnalysisCompleted
-  }
-
   /**
    * Register handlers for the events from the Language Server Protocol that we
    * care about.
    */
   public register(connection: LSP.Connection): void {
-    // The content of a text document has changed. This event is emitted
-    // when the text document first opened or when its content has changed.
+    const hasConfigurationCapability = !!this.clientCapabilities?.workspace?.configuration
+
+    let currentDocument: TextDocument | null = null
+    let initialized = false // Whether the client finished initializing
+
     this.documents.listen(this.connection)
-    this.documents.onDidChangeContent(async (change) => {
-      const { uri } = change.document
 
-      // Load the tree for the modified contents into the analyzer:
-      const analyzeDiagnostics = this.analyzer.analyze(uri, change.document)
-
-      // Run shellcheck diagnostics:
-      let diagnostics: LSP.Diagnostic[] = []
-
-      const folders = this.clientCapabilities.workspace?.workspaceFolders
-        ? await connection.workspace.getWorkspaceFolders()
-        : []
-      const lintDiagnostics = await this.linter.lint(change.document, folders || [])
-      diagnostics = diagnostics.concat(lintDiagnostics)
-
-      // Treesitter's diagnostics can be a bit inaccurate, so we only merge the
-      // analyzer's diagnostics if the setting is enabled:
-      if (config.getHighlightParsingError()) {
-        diagnostics = diagnostics.concat(analyzeDiagnostics)
+    this.documents.onDidChangeContent(({ document }) => {
+      // The content of a text document has changed. This event is emitted
+      // when the text document first opened or when its content has changed.
+      currentDocument = document
+      if (initialized) {
+        this.analyzeAndLintDocument(document)
       }
+    })
 
-      connection.sendDiagnostics({ uri, diagnostics })
+    this.documents.onDidClose((event) => {
+      delete this.uriToCodeActions[event.document.uri]
     })
 
     // Register all the handlers for the LSP events.
@@ -140,10 +127,157 @@ export default class BashServer {
     connection.onReferences(this.onReferences.bind(this))
     connection.onCompletion(this.onCompletion.bind(this))
     connection.onCompletionResolve(this.onCompletionResolve.bind(this))
+    connection.onCodeAction(this.onCodeAction.bind(this))
+
+    /**
+     * The initialized notification is sent from the client to the server after
+     * the client received the result of the initialize request but before the
+     * client is sending any other request or notification to the server.
+     * The server can use the initialized notification for example to dynamically
+     * register capabilities. The initialized notification may only be sent once.
+     */
+    connection.onInitialized(async () => {
+      const { config: environmentConfig, environmentVariablesUsed } =
+        config.getConfigFromEnvironmentVariables()
+
+      if (environmentVariablesUsed.length > 0) {
+        this.updateConfiguration(environmentConfig)
+        connection.console.warn(
+          `Environment variable configuration is being deprecated, please use workspace configuration. The following environment variables were used: ${environmentVariablesUsed.join(
+            ', ',
+          )}`,
+        )
+      }
+
+      if (hasConfigurationCapability) {
+        // Register event for all configuration changes.
+        connection.client.register(LSP.DidChangeConfigurationNotification.type, {
+          section: CONFIGURATION_SECTION,
+        })
+
+        // get current configuration from client
+        const configObject = await connection.workspace.getConfiguration(
+          CONFIGURATION_SECTION,
+        )
+        this.updateConfiguration(configObject)
+        this.connection.console.log('Configuration loaded from client')
+      }
+
+      initialized = true
+      if (currentDocument) {
+        // If we already have a document, analyze it now that we're initialized
+        // and the linter is ready.
+        this.analyzeAndLintDocument(currentDocument)
+      }
+
+      // NOTE: we do not block the server initialization on this background analysis.
+      return { backgroundAnalysisCompleted: this.startBackgroundAnalysis() }
+    })
+
+    // Respond to changes in the configuration.
+    connection.onDidChangeConfiguration(({ settings }) => {
+      const configChanged = this.updateConfiguration(settings[CONFIGURATION_SECTION])
+      if (configChanged && initialized) {
+        this.connection.console.log('Configuration changed')
+        this.startBackgroundAnalysis()
+        if (currentDocument) {
+          this.uriToCodeActions[currentDocument.uri] = undefined
+          this.analyzeAndLintDocument(currentDocument)
+        }
+      }
+    })
 
     // FIXME: re-lint on workspace folder change
   }
 
+  private async startBackgroundAnalysis(): Promise<{ filesParsed: number }> {
+    const { workspaceFolder } = this
+    if (workspaceFolder) {
+      return this.analyzer.initiateBackgroundAnalysis({
+        globPattern: this.config.globPattern,
+        backgroundAnalysisMaxFiles: this.config.backgroundAnalysisMaxFiles,
+      })
+    }
+
+    return Promise.resolve({ filesParsed: 0 })
+  }
+
+  private updateConfiguration(configObject: any): boolean {
+    if (typeof configObject === 'object') {
+      try {
+        const newConfig = config.ConfigSchema.parse(configObject)
+
+        if (!isDeepStrictEqual(this.config, newConfig)) {
+          this.config = newConfig
+
+          // NOTE: I don't really like this... An alternative would be to pass in the
+          // shellcheck executable path when linting. We would need to handle
+          // resetting the canLint flag though.
+
+          const { shellcheckPath } = this.config
+          if (!shellcheckPath) {
+            this.connection.console.log(
+              'ShellCheck linting is disabled as "shellcheckPath" was not set',
+            )
+            this.linter = undefined
+          } else {
+            this.linter = new Linter({
+              console: this.connection.console,
+              executablePath: shellcheckPath,
+            })
+          }
+
+          this.analyzer.setIncludeAllWorkspaceSymbols(
+            this.config.includeAllWorkspaceSymbols,
+          )
+
+          return true
+        }
+      } catch (err) {
+        this.connection.console.warn(`updateConfiguration: failed with ${err}`)
+      }
+    }
+
+    return false
+  }
+
+  /**
+   * Analyze and lint the given document.
+   */
+  public async analyzeAndLintDocument(document: TextDocument) {
+    const { uri } = document
+
+    let diagnostics: LSP.Diagnostic[] = []
+
+    // Load the tree for the modified contents into the analyzer:
+    const analyzeDiagnostics = this.analyzer.analyze({ uri, document })
+    // Treesitter's diagnostics can be a bit inaccurate, so we only merge the
+    // analyzer's diagnostics if the setting is enabled:
+    if (this.config.highlightParsingErrors) {
+      diagnostics = diagnostics.concat(analyzeDiagnostics)
+    }
+
+    // Run ShellCheck diagnostics:
+    if (this.linter) {
+      try {
+        const folders = this.clientCapabilities.workspace?.workspaceFolders
+          ? await this.connection.workspace.getWorkspaceFolders()
+          : []
+        const { diagnostics: lintDiagnostics, codeActions } = await this.linter.lint(
+          document,
+          folders || [],
+          this.config.shellcheckArguments,
+        )
+        diagnostics = diagnostics.concat(lintDiagnostics)
+        this.uriToCodeActions[uri] = codeActions
+      } catch (err) {
+        this.connection.console.error(`Error while linting: ${err}`)
+      }
+    }
+
+    this.connection.sendDiagnostics({ uri, version: document.version, diagnostics })
+  }
+
   /**
    * The parts of the Language Server Protocol that we are currently supporting.
    */
@@ -162,6 +296,11 @@ export default class BashServer {
       documentSymbolProvider: true,
       workspaceSymbolProvider: true,
       referencesProvider: true,
+      codeActionProvider: {
+        codeActionKinds: [LSP.CodeActionKind.QuickFix],
+        resolveProvider: false,
+        workDoneProgress: false,
+      },
     }
   }
 
@@ -206,19 +345,24 @@ export default class BashServer {
   }: {
     symbol: LSP.SymbolInformation
     currentUri: string
-  }): string {
+  }): LSP.MarkupContent {
     const symbolUri = symbol.location.uri
-    const symbolStarLine = symbol.location.range.start.line
+    const symbolStartLine = symbol.location.range.start.line
 
-    const commentAboveSymbol = this.analyzer.commentsAbove(symbolUri, symbolStarLine)
+    const commentAboveSymbol = this.analyzer.commentsAbove(symbolUri, symbolStartLine)
     const symbolDocumentation = commentAboveSymbol ? `\n\n${commentAboveSymbol}` : ''
-    const hoverHeader = `### ${symbolKindToDescription(symbol.kind)}: **${symbol.name}**`
+    const hoverHeader = `${symbolKindToDescription(symbol.kind)}: **${symbol.name}**`
     const symbolLocation =
       symbolUri !== currentUri
-        ? `in ${Path.relative(currentUri, symbolUri)}`
-        : `on line ${symbolStarLine + 1}`
+        ? `in ${path.relative(path.dirname(currentUri), symbolUri)}`
+        : `on line ${symbolStartLine + 1}`
 
-    return `${hoverHeader} - *defined ${symbolLocation}*${symbolDocumentation}`
+    // TODO: An improvement could be to add show the symbol definition in the hover instead
+    // of the defined location – similar to how VSCode works for languages like TypeScript.
+
+    return getMarkdownContent(
+      `${hoverHeader} - *defined ${symbolLocation}*${symbolDocumentation}`,
+    )
   }
 
   private getCompletionItemsForSymbols({
@@ -259,7 +403,7 @@ export default class BashServer {
       return null
     }
 
-    const explainshellEndpoint = config.getExplainshellEndpoint()
+    const { explainshellEndpoint } = this.config
     if (explainshellEndpoint) {
       try {
         const { helpHTML } = await this.analyzer.getExplainshellDocumentation({
@@ -290,13 +434,13 @@ export default class BashServer {
     ) {
       const shellDocumentation = await getShellDocumentation({ word })
       if (shellDocumentation) {
-        // eslint-disable-next-line no-console
-        return { contents: getMarkdownContent(shellDocumentation) }
+        return { contents: getMarkdownContent(shellDocumentation, 'man') }
       }
     } else {
       const symbolDocumentation = deduplicateSymbols({
         symbols: this.analyzer.findSymbolsMatchingWord({
           exactMatch: true,
+          uri: currentUri,
           word,
         }),
         currentUri,
@@ -321,7 +465,11 @@ export default class BashServer {
     if (!word) {
       return null
     }
-    return this.analyzer.findDefinition({ word })
+    return this.analyzer.findDefinition({
+      position: params.position,
+      uri: params.textDocument.uri,
+      word,
+    })
   }
 
   private onDocumentSymbol(params: LSP.DocumentSymbolParams): LSP.SymbolInformation[] {
@@ -410,9 +558,10 @@ export default class BashServer {
         ? []
         : this.getCompletionItemsForSymbols({
             symbols: shouldCompleteOnVariables
-              ? this.analyzer.getAllVariableSymbols()
+              ? this.analyzer.getAllVariableSymbols({ uri: currentUri })
               : this.analyzer.findSymbolsMatchingWord({
                   exactMatch: false,
+                  uri: currentUri,
                   word,
                 }),
             currentUri,
@@ -483,6 +632,46 @@ export default class BashServer {
     return allCompletions
   }
 
+  private async onCodeAction(params: LSP.CodeActionParams): Promise<LSP.CodeAction[]> {
+    const codeActions = this.uriToCodeActions[params.textDocument.uri]
+
+    if (!codeActions) {
+      return []
+    }
+
+    const getDiagnosticsFingerPrint = (diagnostics?: LSP.Diagnostic[]): string[] =>
+      (diagnostics &&
+        diagnostics
+          .map(({ code, source, range }) =>
+            code !== undefined && source && range
+              ? JSON.stringify({
+                  code,
+                  source,
+                  range,
+                })
+              : null,
+          )
+          .filter((fingerPrint): fingerPrint is string => fingerPrint != null)) ||
+      []
+
+    const paramsDiagnosticsKeys = getDiagnosticsFingerPrint(params.context.diagnostics)
+
+    // find actions that match the paramsDiagnosticsKeys
+    const actions = codeActions.filter((action) => {
+      const actionDiagnosticsKeys = getDiagnosticsFingerPrint(action.diagnostics)
+      // actions without diagnostics are always returned
+      if (actionDiagnosticsKeys.length === 0) {
+        return true
+      }
+
+      return actionDiagnosticsKeys.some((actionDiagnosticKey) =>
+        paramsDiagnosticsKeys.includes(actionDiagnosticKey),
+      )
+    })
+
+    return actions
+  }
+
   private async onCompletionResolve(
     item: LSP.CompletionItem,
   ): Promise<LSP.CompletionItem> {
@@ -506,7 +695,7 @@ export default class BashServer {
       return documentation
         ? {
             ...item,
-            documentation: getMarkdownContent(documentation),
+            documentation: getMarkdownContent(documentation, 'man'),
           }
         : item
     } catch (error) {
@@ -609,18 +798,19 @@ function symbolKindToDescription(s: LSP.SymbolKind): string {
   }
 }
 
-const getMarkdownContent = (documentation: string): LSP.MarkupContent => ({
-  value: ['``` man', documentation, '```'].join('\n'),
-  // Passed as markdown for syntax highlighting
-  kind: 'markdown' as const,
-})
+function getMarkdownContent(documentation: string, language?: string): LSP.MarkupContent {
+  return {
+    value: language
+      ? // eslint-disable-next-line prefer-template
+        ['``` ' + language, documentation, '```'].join('\n')
+      : documentation,
+    kind: LSP.MarkupKind.Markdown,
+  }
+}
 
 function getCommandOptions(name: string, word: string): string[] {
   // TODO: The options could be cached.
-  const options = Process.spawnSync(Path.join(__dirname, '../src/get-options.sh'), [
-    name,
-    word,
-  ])
+  const options = spawnSync(path.join(__dirname, '../src/get-options.sh'), [name, word])
 
   if (options.status !== 0) {
     return []
